/**
   Shrink each line segment from each model by the
   amount alpha. (If alpha > 1, this will actually
   stretch each line segment.)

   This class is a "pipeline stage" because it
   extends the PipelineStage abstract class, which
   consists of a single method, processModel().
*/

public class ShrinkShader extends PipelineStage
{
   public double alpha;

   /**
      This constructor initializes this shader's
      reference to the SceneRender object that
      contains it. This constructor also sets
      a parameter needed by this shader. Alpha
      should be a small positive number (alpha < 1.0).
   */
   public ShrinkShader(double alpha, SceneRender sceneRender)
   {
      super(sceneRender);

      this.alpha = alpha;
   }


   /**
      Shorten, by the amount alpha, the length of each line segment from this model's list of line segments.
   */
   public Model processModel(Model model)
   {
      for (LineSegment ls : model.lineSegmentList)
      {
         Vertex v0 = ls.v[0];
         Vertex v1 = ls.v[1];

         double new_v0_x = (1 - alpha/2.0) * v0.x + alpha/2.0 * v1.x;
         double new_v0_y = (1 - alpha/2.0) * v0.y + alpha/2.0 * v1.y;
         double new_v0_z = (1 - alpha/2.0) * v0.z + alpha/2.0 * v1.z;
         double new_v1_x = (1 - alpha/2.0) * v1.x + alpha/2.0 * v0.x;
         double new_v1_y = (1 - alpha/2.0) * v1.y + alpha/2.0 * v0.y;
         double new_v1_z = (1 - alpha/2.0) * v1.z + alpha/2.0 * v0.z;

         double new_v0_r = (1 - alpha/2.0) * v0.r + alpha/2.0 * v1.r;
         double new_v0_g = (1 - alpha/2.0) * v0.g + alpha/2.0 * v1.g;
         double new_v0_b = (1 - alpha/2.0) * v0.b + alpha/2.0 * v1.b;
         double new_v1_r = (1 - alpha/2.0) * v1.r + alpha/2.0 * v0.r;
         double new_v1_g = (1 - alpha/2.0) * v1.g + alpha/2.0 * v0.g;
         double new_v1_b = (1 - alpha/2.0) * v1.b + alpha/2.0 * v0.b;

         v0.x = new_v0_x;
         v0.y = new_v0_y;
         v0.z = new_v0_z;
         v1.x = new_v1_x;
         v1.y = new_v1_y;
         v1.z = new_v1_z;

         v0.r = new_v0_r;
         v0.g = new_v0_g;
         v0.b = new_v0_b;
         v1.r = new_v1_r;
         v1.g = new_v1_g;
         v1.b = new_v1_b;
      }

      return model;
   }
}