/*
   The program creates a simple nested model with
   the following nested structure.

               m1
              /  \
            /      \
           m2       m3
            \      /
              \  /
               m4

   The model m4 at the bottom of the DAG (directed
   acyclic graph) is the square with vertices at
   +1 and -1 which is then rotated 45 degrees.

   The model m2 translates m4 by sqrt(2) to the left (so one
   corner of the rotated square is at the origin).

   The model m3 translates m4 by sqrt(2) to the right (so one
   corner of the rotated square is again at the origin).

   To see how the "zeroth" stage of the rendering pipeline,
   P0_CopyModel.java, manipulates the above DAG into a tree,
   run this program with the environment variable PipeLine0
   set to 1 (use the Run.cmd file to do this).
*/
import java.awt.Color;

public class VerySimpleNestedModel
{
   public static void main(String[] args)
   {
      // Create a Model object to hold the geometry.
      Model m4 = new Model();

      // Add geometry to the Model.
      Vertex v1 = new Vertex( 1.0,  1.0, 0.0);
      Vertex v2 = new Vertex(-1.0,  1.0, 0.0);
      Vertex v3 = new Vertex(-1.0, -1.0, 0.0);
      Vertex v4 = new Vertex( 1.0, -1.0, 0.0);
      // square.
      m4.lineSegmentList.add( new LineSegment(v1, v2) );
      m4.lineSegmentList.add( new LineSegment(v2, v3) );
      m4.lineSegmentList.add( new LineSegment(v3, v4) );
      m4.lineSegmentList.add( new LineSegment(v4, v1) );

      m4.modelRotate(45, 0, 0, 1); // rotate 45 degrees around z-axis

      // Give the Model a color.
      m4.setColor(Color.blue);

      // Make a translated copy of m4
      Model m2 = new Model();
      m2.addSubModel(m4);
      m2.modelTranslate(-Math.sqrt(2), 0, 0);

      // Make another translated copy of m4
      Model m3 = new Model();
      m3.addSubModel(m4);
      m3.modelTranslate(Math.sqrt(2), 0, 0);

      // Create our "top level" model containing the submodels.
      Model m1 = new Model();
      m1.addSubModel(m2);
      m1.addSubModel(m3);

      // Create the Scene object that we shall render
      Scene scene = new Scene();

      // Add the top level model to the Scene.
      scene.modelList.add(m1);

      // Create a Camera object.
      Camera camera = new Camera();

      // Give the Camera a view matrix.
      camera.view2Identity();

      // Set up the camera's view frustum.
      double left   = -5.0;
      double right  =  5.0;
      double top    =  5.0;
      double bottom = -5.0;
      double near   =  1.0;
      double far    =  20.0;
      camera.projOrtho(left, right, bottom, top, near, far);

      // Set up the camera's location and orientation.
      double eyex=0,    eyey=0,    eyez=10;
      double centerx=0, centery=0, centerz=0;
      double upx=0,     upy=1,     upz=0;
      camera.viewLookAt(eyex,    eyey,    eyez,
                        centerx, centery, centerz,
                        upx,     upy,     upz);

      // Add the Camera to the Scene.
      scene.setCamera(camera);

      System.err.println( m1.toString() );

      // Create a framebuffer to render our scene into.
      int vp_width  = 512;
      int vp_height = 512;
      FrameBuffer fb = new FrameBuffer(vp_width, vp_height);
      // Give the framebuffer a nice background color.
      fb.clearFB(java.awt.Color.lightGray);
      // Create a default rendering pipeline.
      SceneRender renderer = new SceneRender();
      // Render our scene into the frame buffer.
      renderer.renderScene(scene, fb);
      // Save the resulting image in a file.
      fb.dumpFB2File( String.format("PPM_VerySimpleNestedModel_Frame%02d.ppm", 0) );

      // Spin the model 360 degrees.
      for (int i = 1; i <= 36; i++)
      {
         // Rotate the model by 10 degrees.
         m1.modelRotate(10, 0, 1, 0);
         m1.modelRotate(10, 1, 0, 0);
         m1.modelRotate(10, 0, 0, 1);
         // Render again.
         fb.clearFB(java.awt.Color.lightGray);
         renderer.renderScene(scene, fb);
         fb.dumpFB2File(String.format("PPM_VerySimpleNestedModel_Frame%02d.ppm", i));
      }

   }
}