/**
   A Vertex object has four doubles which represent the (homogeneous) coordinates of a point in 3-dimensional space. The fourth, homogeneous, coordinate will usually be 1, but in some stages of the graphics rendering pipeline it will be some other (non-zero) number.

   When a Vertex object is created in a client program, before the Vertex object moves down the graphics  rendering pipeline, the coordinates in the Vertex will be in "world coordinates".

   As a Vertex object moves down the graphics rendering pipeline, the coordinates in the Vertex will be transformed from one coordinate system to another (from world to view to clip to viewport coordinates).
*/
import java.util.Random;
import java.awt.Color;

public class Vertex
{
   public double x, y, z, w;           // a vertex in homogenous coordinates
   public double r=0.0, g=0.0, b=0.0;  // color at this vertex

   public Vertex()
   {
      this(0.0, 0.0, 0.0, 1.0);
   }


   public Vertex(double x, double y, double z)
   {
      this(x, y, z, 1.0);
   }


   public Vertex(float[] v)
   {
      this(v[0], v[1], v[2], v[3]);
   }


   public Vertex(double x, double y, double z, double w)
   {
      set(x, y, z, w);
   }


   public Vertex(Vertex v) // a "copy constructor"
   {
      set(v.x, v.y, v.z, v.w);
      setColor(v);
   }


   private void set(double x, double y, double z)
   {
      set(x, y, z, 1.0);
   }


   private void set(double x, double y, double z, double w)
   {
      this.x = x;
      this.y = y;
      this.z = z;
      this.w = w;
   }


   /**
      Set the color of this vertex to (r,g,b).
   */
   public void setColor(double r, double g, double b)
   {
      this.r = r;
      this.g = g;
      this.b = b;

      if ( (r < 0.0) || (r > 1.0)
        || (g < 0.0) || (g > 1.0)
        || (b < 0.0) || (b > 1.0) )
      {
         System.err.println("ERROR! Invalid double color for vertex");
         System.err.println( this.toString(3) );
         System.exit(-1);
      }
   }


   public void setColor(int r, int g, int b)
   {
      this.r = ((float)r)/(float)255;
      this.g = ((float)g)/(float)255;
      this.b = ((float)b)/(float)255;

      if ( (r < 0) || (r > 255)
        || (g < 0) || (g > 255)
        || (b < 0) || (b > 255))
      {
         System.err.println("ERROR! Invalid int color for vertex");
         System.err.println( this.toString(3) );
         System.exit(-1);
      }
   }


   public void setColor(Color c)
   {
      setColor(c.getRed(), c.getGreen(), c.getBlue());
   }


   public void setColor(Vertex v) // copy the color from v to this vertex
   {
      setColor(v.r, v.g, v.b);
   }


   public void setColorRandom()
   {
      Random generator = new Random();
      double r = generator.nextDouble();
      double g = generator.nextDouble();
      double b = generator.nextDouble();
      setColor(r, g, b);
   }


   public Color getColor()
   {
      return new Color((float)this.r, (float)this.g, (float)this.b);
   }


   /**
      Vertex plus Vector returns a Vertex.
      This can be used to translate a Vertex (but a translation transformation would be better).
   */
   public Vertex plus(Vector v)
   {
      Vertex temp = new Vertex( this.x+(v.x), this.y+(v.y), this.z+(v.z), this.w+(v.w) );
      temp.setColor(this);
      return temp;
   }


   /**
      Vertex minus Vertex returns a Vector.
      Used to find the displacement vector between two vertices.
   */
   public Vector minus(Vertex v)
   {
      return new Vector( this.x-(v.x), this.y-(v.y), this.z-(v.z), this.w-(v.w) );
   }


   /**
      Scalar times Vertex returns a Vertex.
      Needed for the perspective division step in the rendering pipeline.
   */
   public Vertex times(double scalar)
   {
      Vertex temp =  new Vertex(scalar*this.x, scalar*this.y, scalar*this.z, scalar*this.w);
      temp.setColor(this);
      return temp;
   }


   /**
      For debugging.
   */
   public String toString()
   {
      // Here is one way to get programable precision and width.
      int p = 5;     // the precision for the following format string
      int t = p + 4; // the width for the folowing format string
      String format =      "(x,y,z,w) = (% "+t+"."+p+"f  % "+t+"."+p+"f  % "+t+"."+p+"f  % "+t+"."+p+"f)\n";
      return String.format( format, x, y, z, w);
    //return String.format("(x,y,z,w) = (% .5f  % .5f  % .5f  % .5f)", x, y, z, w);
   }


   /**
      For debugging. What this method prints out
      depends on the value of the parameter vertexFields.

      The value of vertexFields is treated as a bit field.
      The bits have the following meanings.
        1 => print vertex coordinates
        2 => print r g b values
   */
   public String toString(int vertexFields)
   {
      String result = "";
      int p = 5;     // the precision for the following format string
      int t = p + 4; // the width for the folowing format string
      if (1 == (vertexFields & 1))
      {
         // Here is one way to get programable precision and width.
         String format = "(x,y,z,w) = (% "+t+"."+p+"f  % "+t+"."+p+"f  % "+t+"."+p+"f  % "+t+"."+p+"f)\n";
         result += String.format( format, x, y, z, w);
      }
      if (2 == (vertexFields & 2))
      {
         result += String.format("<r,g,b> = <% .5f  % .5f  % .5f>\n", r, g, b);
      }
      return result;
   }
}