/**
   We need to make a copy of this model so that future
   pipeline stages can mutate the copy as they render it.
   (If we let the renderer mutate our original copy of this
   model, then we would not be able to re-render this model,
   say from a different view point.)

   Since we now have nested models, we have more work to do
   in this stage than in the previous renderer. We need to do
   three things to each model.

   1) Make a copy of this model so that future pipeline
   stages can mutate the copy as they render it. In
   addition, traverse this model's DAG of nested models
   and recursively copy them. This will convert the DAG
   (directed acyclic graph) of nested models into a tree
   of nested models. For example, if this model has the
   following DAG structure,

               m1
              /  \
            /      \
           m2       m3
            \      /
              \  /
               m4

   then after the first step of this stage we will have
   the following tree structure.

               m1'
              /  \
            /      \
          m2'       m3'
          |         |
          |         |
          m4'       m4"

   where m1', m2', m3' are copies of m1, m2, m3, and
   m4' and m4" are both copies of m4.

   2) Transform all the vertices in the nested models
   into a single model coordinate system.

   3) Flatten the line segment lists from the nested models
   into one single line segment list in the top level model.
   (We only need to do this if we want to keep the rest of
   the pipeline the same as it was before we added nested
   models. We could instead rewrite the other stages to
   traverse the tree of nested models.)


   NOTE: This copying stage is analogous to the need for
   the CPU of a computer to copy all of the Scene data
   across a hardware bus to the computer's graphics card
   (the GPU).
*/
import java.util.List;

public class P0_CopyModel extends PipelineStage
{
   /**
      This constructor initializes this object's
      reference to the SceneRender object that
      contains it. This gives this stage access
      to addditional information that might be
      needed to process the Models passed to the
      processModel() method.
   */
   public P0_CopyModel(SceneRender sceneRender)
   {
      super(sceneRender);
   }


   /**
      Create a new Model object that contains copies
      of all of the line segments from this model.
   */
   public Model processModel(Model model)
   {
      // Create a new Model that is a copy of the original Model.
      // This copied model has the same nested model structure as
      // the original model, but it is a Tree instead of a DAG.
      Model newModel = copy( model );

      // For each sub-model of the new Model, transform all of its
      // vertices to one model coordinate system. This transformed
      // model still has the same nested model structure as the
      // original model but it has all of its line segments in the
      // same model coordinate system.
      for (Model nestedModel : newModel.nestedModels)
      {
         transform( nestedModel, new Matrix(1.0) );
      }

      // For each sub-model of the new model, copy all of its line
      // segments into the line segment list at the top level.
      for (Model nestedModel : newModel.nestedModels)
      {
         flatten( nestedModel, newModel.lineSegmentList );
      }

      // We no longer need the list of sub models in the new model since
      // everything has been moved into the top level of the new model.
      newModel.nestedModels.clear();

      return newModel;
   }


   /**
      This is essentially a pre-order traversal of the tree of nested models.
   */
   public Model copy(Model model)
   {
      Model newModel = new Model(model); // copy everything except the geometry

      // Copy all the line segments in the model's line segment list.
      for (LineSegment ls : model.lineSegmentList)
      {
         LineSegment lsCopy = new LineSegment(ls);
         newModel.addLineSegment( lsCopy );
      }

      // Recursively copy all the line segments in the sub-models.
      for (Model nestedModel : model.nestedModels)
      {
         newModel.addSubModel( copy(nestedModel) );
      }

      return newModel;
   }//copy()


   /**
      This is also a pre-order traversal of the tree of nested models.
   */
   public void transform(Model model, Matrix ctm)
   {
      ctm = ctm.times( model.modelMatrix ); // ctm is the "current transformation matrix"

      // Transform all the line segments in the model's line segment list.
      for (LineSegment ls : model.lineSegmentList)
      {
         // transform the vertices
         ls.v[0] = ctm.times(ls.v[0]);
         ls.v[1] = ctm.times(ls.v[1]);
      }

      // Recursively transform all the line segments in the sub-models.
      for (Model nestedModel : model.nestedModels)
      {
         transform(nestedModel, ctm);
      }
   }//transform()


   /**
      This is also a pre-order traversal of the tree of nested models.
   */
   private void flatten(Model model, List<LineSegment> topLevelLineSegmentList)
   {
      // Flatten all the line segments in this model's line segment list.
      topLevelLineSegmentList.addAll( model.lineSegmentList );

      // Recursively flatten all the sub-models.
      for (Model nestedModel : model.nestedModels)
      {
         flatten(nestedModel, topLevelLineSegmentList);
      }
   }//flatten()
}