/**
   Rasterize each line segment into pixels in the framebuffer.
*/
import java.awt.Color;
import java.util.LinkedList;

public class P6_Rasterize extends PipelineStage
{
   /**
      This constructor initializes this object's
      reference to the SceneRender object that
      contains it. This gives this stage access
      to addditional information that might be
      needed to process the Models passed to the
      processModel() method.
   */
   public P6_Rasterize(SceneRender sceneRender)
   {
      super(sceneRender);
   }


   /**
      Walk the list of LineSegment objects and rasterize each line segment.

      This rasterization algorithm is from
        "Fundamentals of Computer Graphics", 3rd Edition,
         by Peter Shirley, pages 163-165.
   */
   public Model processModel(Model model)
   {
      FrameBuffer fb = this.sceneRender.fb;

      for (LineSegment ls : model.lineSegmentList)
      {
         int x0 = (int)ls.v[0].x;
         int y0 = (int)ls.v[0].y;
         int x1 = (int)ls.v[1].x;
         int y1 = (int)ls.v[1].y;
         if (x0 == fb.getWidthVP())  x0--;
         if (y0 == fb.getHeightVP()) y0--;
         if (x1 == fb.getWidthVP())  x1--;
         if (y1 == fb.getHeightVP()) y1--;

         // get the colors from the two vertices
         double r0 = ls.v[0].r;
         double g0 = ls.v[0].g;
         double b0 = ls.v[0].b;
         double r1 = ls.v[1].r;
         double g1 = ls.v[1].g;
         double b1 = ls.v[0].b;

         if (Math.abs(y1 - y0) <= Math.abs(x1 - x0)) // if abs(slope) <= 1
         {
            if (x1 < x0) // swap (x0, y0) with (x1, y1)
            {
               int tempX = x0;
               int tempY = y0;
               x0 = x1;
               y0 = y1;
               x1 = tempX;
               y1 = tempY;
               // swap the colors too
               double tempR = r0;
               double tempG = g0;
               double tempB = b0;
               r0 = r1;
               g0 = g1;
               b0 = b1;
               r1 = tempR;
               g1 = tempG;
               b1 = tempB;
            }

            double slope  = (double)(y1 - y0) / (double)(x1 - x0);
            double slopeR = (double)(r1 - r0) / (double)(x1 - x0);
            double slopeG = (double)(g1 - g0) / (double)(x1 - x0);
            double slopeB = (double)(b1 - b0) / (double)(x1 - x0);

            for (int x = x0; x <= x1; x++)
            {
               int y = (int)Math.round( y0 + slope*(x - x0) );
               if (y == fb.height) y--;
               // interpolate this pixel's color between the two endpoint's colors
               float r = (float)Math.abs(r0 + slopeR*(x - x0));
               float g = (float)Math.abs(g0 + slopeG*(x - x0));
               float b = (float)Math.abs(b0 + slopeB*(x - x0));
               // we need the Math.abs() because otherwise, we sometimes get -0.0

             //System.out.printf("x=%d, y=%d <r=%f, g=%f, b=%f>\n", x, y, r, g, b);

               // set this pixel in the framebuffer
               fb.setPixelVP(x, y, new Color(r, g, b));
            }
         }
         else  // abs(slope) > 1
         {
            if (y1 < y0) // swap (x0, y0) with (x1, y1)
            {
               int tempX = x0;
               int tempY = y0;
               x0 = x1;
               y0 = y1;
               x1 = tempX;
               y1 = tempY;
               // swap the colors too
               double tempR = r0;
               double tempG = g0;
               double tempB = b0;
               r0 = r1;
               g0 = g1;
               b0 = b1;
               r1 = tempR;
               g1 = tempG;
               b1 = tempB;
            }

            double slope  = (double)(x1 - x0) / (double)(y1 - y0);
            double slopeR = (double)(r1 - r0) / (double)(y1 - y0);
            double slopeG = (double)(g1 - g0) / (double)(y1 - y0);
            double slopeB = (double)(b1 - b0) / (double)(y1 - y0);

            for (int y = y0; y <= y1; y++)
            {
               int x = (int)Math.round( x0 + slope*(y - y0) );
               if (x == fb.width)  x--;
               // interpolate this pixel's color between the two endpoint's colors
               float r = (float)Math.abs(r0 + slopeR*(y - y0));
               float g = (float)Math.abs(g0 + slopeG*(y - y0));
               float b = (float)Math.abs(b0 + slopeB*(y - y0));
               // we need the Math.abs() because otherwise, we sometimes get -0.0

             //System.out.printf("x=%d, y=%d <r=%f, g=%f, b=%f>\n", x, y, r, g, b);

               // set this pixel in the framebuffer
               fb.setPixelVP(x, y, new Color(r, g, b));
            }
         }
      }
      return model;
   }
}