/**
   This is a right, triangular prism with length 2h, side length 1, centered
   at the origin, and with its central axis on the y-axis.
   http://en.wikipedia.org/wiki/Prism_(geometry)

   Attached to each of the two triangular ends of the prism is a tetrahedron
   of height h2. So the total height is 2*(h + h2).
*/

public class PrismModel extends Model
{
   /**
      A prism with a regular tetrahedron at the top and bottom.
   */
   public PrismModel(double h)
   {
      this(h, (float)(Math.sqrt(2)/Math.sqrt(3)) );
   }


   /**
      A prism with a tetrahedron at the top and bottom, where
      the face-edge-face angle of the tetrahedron is theta.
      theta is in radians!
      (If theta = arctan(2*sqrt(2)), then the tetrahedron is regular.)
   */
   public PrismModel(double h, double theta)
   {
      this(h, (float)(Math.tan(theta)/(2*Math.sqrt(3))) );
   }


   /**
      A prism of length h with a tetrahedron of height h2 at the top and bottom.
      NOTE: h2 is a float so that we can distinguish this constructor from the previous one.
   */
   private PrismModel(double h, float h2)
   {
      super();

      // Create the prism's geometry.
      double sqrt3 = Math.sqrt(3.0);
      double sqrt2 = Math.sqrt(2.0);
      Vertex[] v = new Vertex[8];

      v[0] = new Vertex( 1.0/sqrt3,  h,  0);    // three vertices around the top
      v[1] = new Vertex(-0.5/sqrt3,  h,  0.5);
      v[2] = new Vertex(-0.5/sqrt3,  h, -0.5);

      v[3] = new Vertex( 1.0/sqrt3, -h,  0);    // three vertices around the bottom
      v[4] = new Vertex(-0.5/sqrt3, -h,  0.5);
      v[5] = new Vertex(-0.5/sqrt3, -h, -0.5);

      v[6] = new Vertex( 0,  h+h2, 0);  // vertex at the top
      v[7] = new Vertex( 0, -h-h2, 0);  // vertex at the bottom

      // Create 15 line segments.
      LineSegment[] ls = new LineSegment[15];
      int i = 0;
      // 3 top faces
      ls[i++] = new LineSegment(v[6], v[0]);
      ls[i++] = new LineSegment(v[6], v[1]);
      ls[i++] = new LineSegment(v[6], v[2]);
      // the top edge
      ls[i++] = new LineSegment(v[0], v[1]);
      ls[i++] = new LineSegment(v[1], v[2]);
      ls[i++] = new LineSegment(v[2], v[0]);
      // three vertical edges
      ls[i++] = new LineSegment(v[0], v[3]);
      ls[i++] = new LineSegment(v[1], v[4]);
      ls[i++] = new LineSegment(v[2], v[5]);
      // the bottom edge
      ls[i++] = new LineSegment(v[3], v[4]);
      ls[i++] = new LineSegment(v[4], v[5]);
      ls[i++] = new LineSegment(v[5], v[3]);
      // 3 bottom faces
      ls[i++] = new LineSegment(v[7], v[3]);
      ls[i++] = new LineSegment(v[7], v[4]);
      ls[i++] = new LineSegment(v[7], v[5]);

      // Add the line segments to the model.
      for (i = 0; i < ls.length; i++)
      {
         this.addLineSegment( ls[i] );
      }
   }
}