/*
   Here is the "scene graph" for this example.
   Only the PrismModel node holds any geometry.
   All of the other nodes hold only a matrix.

          LinkedModels
            /    \
           /       \
        link1      link2
           \       /
            \    /
              \/
             mTop
            /|  |\
          /  |  |  \
        /    |  |    \
      m1    m2  m3    m4
        \    |  |    /
          \  |  |  /
            \|  |/
              m0
              |
              |
          PrismModel
*/
import java.awt.Color;

public class NestedPrisms_v2
{
   public static void main(String[] args)
   {
      double sqrt3 = Math.sqrt(3.0);

      // Create a Prism object.
      Model m0 = new PrismModel(2, Math.PI/4.0);


      // Create four new Prism objects from m0.
      // right
      Model m1 = new Model();
      m1.addSubModel( m0 );
      m1.modelTranslate(2+0.5/sqrt3, 0, 0);
      // left
      Model m2 = new Model();
      m2.addSubModel( m0 );
      m2.modelTranslate(-2-0.5/sqrt3, 0, 0);
      m2.modelRotate(180, 0,0,1);
      // top
      Model m3 = new Model();
      m3.addSubModel( m0 );
      m3.modelRotate(90, 0,0,1);
      m3.modelTranslate(2+0.5/sqrt3, 0, 0);
      // bottom
      Model m4 = new Model();
      m4.addSubModel( m0 );
      m4.modelRotate(-90, 0,0,1);
      m4.modelTranslate(2+0.5/sqrt3, 0, 0);

      // Creat a top level model to hold the four sub models
      Model mTop = new Model();
      mTop.addSubModel(m1);
      mTop.addSubModel(m2);
      mTop.addSubModel(m3);
      mTop.addSubModel(m4);

      // Create two "linked" copies of the top level model.
      Model link1 = new Model();
      link1.addSubModel(mTop);
      Model link2 = new Model();
      link2.addSubModel(mTop);
      link2.modelTranslate(2, 0, 0);
      link2.modelRotate(90, 1,0,0);
      Model linkedModels = new Model();
      linkedModels.addSubModel(link1);
      linkedModels.addSubModel(link2);


      Model floor = new PanelModel(-5, 5, -5, 5);
      floor.modelTranslate(0, -4, 0);


      // Create the Scene object that we shall render
      Scene scene = new Scene();

      // Add the top level model to the Scene.
      scene.modelList.add(linkedModels);
      scene.modelList.add(floor);

      // Create a Camera object.
      Camera camera = new Camera();

      // Give the Camera a view matrix.
      camera.view2Identity();

      // Set up the camera's view frustum.
      double fov    = 60.0;
      double aspect = 1.0;
      double near   = 1.0;
      double far    = 20.0;
      camera.projPerspective(fov, aspect, near, far);

      // Set up the scene's "camera".
      double eyex=0,    eyey=0,    eyez=10;
      double centerx=0, centery=0, centerz=0;
      double upx=0,     upy=1,     upz=0;
      camera.viewLookAt(eyex,    eyey,    eyez,
                        centerx, centery, centerz,
                        upx,     upy,     upz);

      // Add the Camera to the Scene.
      scene.setCamera(camera);

      // Create a framebuffer to render our scene into.
      int vp_width  = 512;
      int vp_height = 512;
      FrameBuffer fb = new FrameBuffer(vp_width, vp_height);
      // Give the framebuffer a nice background color.
      fb.clearFB(java.awt.Color.lightGray);
      // Create a default rendering pipeline,
      SceneRender renderer = new SceneRender();

      // Render our scene into the frame buffer.
      renderer.renderScene(scene, fb);
      // Save the resulting image in a file.
      String basename = "PPM_NestedPrism_v2_Frame_";
      fb.dumpFB2File( String.format(basename + "%02d.ppm", 0) );

      // Spin the model 360 degrees arond two axes.
      for (int i = 1; i <= 72; i++)
      {
         // Rotate the top level model 5 degrees.
         linkedModels.model2Identity();
         linkedModels.modelRotate( 5*i, 1,0,0);
         linkedModels.modelRotate( 5*i, 0,1,0);
         linkedModels.modelRotate( 5*i, 0,0,1);

         // Render again.
         fb.clearFB(java.awt.Color.lightGray);
         renderer.renderScene(scene, fb);
         fb.dumpFB2File(String.format(basename + "%02d.ppm", i));
      }
   }
}